# -*- coding: utf-8 -*-

import logging


class PrintHostTags:
    @staticmethod
    def run(data, **kwargs):
        for host_name in data.targets["hosts"]:
            message = "HOST TAG LOOKUP \n %s tags:\n\t%s"
            if host_name in data.data["hosts"]:
                host_tags = data.data["hosts"][host_name]
                logging.info(message % (host_name, str(host_tags)))
        return data


class PrintHostApplications:
    @staticmethod
    def run(data, **kwargs):
        message = "HOST TO APPLICATION LOOKUP \n HOST: %s has applications:\n\t%s"
        host_name = kwargs.get("host_name")
        logging.info(message % (host_name, data.targets["applications"]))
        return data


class PrintTagHosts:
    @staticmethod
    def run(data, **kwargs):
        message = "TAG TO HOST LOOKUP \n TAG: %s=%s has hosts:\n\t%s"
        tag_name = kwargs.get("tag_name")
        tag_value = kwargs.get("tag_value")
        logging.info(message % (tag_name, tag_value, list(data.targets["hosts"])))
        return data


class PrintTagApplications:
    @staticmethod
    def run(data, **kwargs):
        message = "TAG TO APPLICATION LOOKUP \n TAG: %s=%s has applications:\n\t%s"
        tag_name = kwargs.get("tag_name")
        tag_value = kwargs.get("tag_value")
        logging.info(message % (tag_name, tag_value, data.targets["applications"]))
        return data


class PrintApplicationHosts:
    @staticmethod
    def run(data, **kwargs):
        message = "APPLICATION TO HOST LOOKUP \n APPLICATION: %s has hosts:\n\t%s"
        application_name = kwargs.get("application_name")
        logging.info(message % (application_name, data.targets["hosts"]))
        return data


class PrintApplicationTags:
    @staticmethod
    def run(data, **kwargs):
        message = "APPLICATION TO TAG LOOKUP \n APPLICATION: %s has tags:\n\t%s"
        application_name = kwargs.get("application_name")
        logging.info(message % (application_name, data.data["application_hosts"][application_name]))
        return data


class PrintApplicationProperties:
    @staticmethod
    def run(data, **kwargs):
        message = "APPLICATION TO PROPERTIES LOOKUP \n APPLICATION: %s has properties:\n\t%s"
        appliation_name = kwargs.get("application_name")
        logging.info(message % (appliation_name, list(data.data["application_properties"][appliation_name]["properties"])))
        return data


class PrintPropertyApplications:
    @staticmethod
    def run(data, **kwargs):
        message = "PROPERTIES TO APPLICATION LOOKUP \n PROPERTIES: %s has applications:\n\t%s"
        property_name = kwargs.get("property_name")
        logging.info(message % (property_name, data.targets["applications"]))
        return data


class PrintHierarchy:
    @staticmethod
    def run(data, **kwargs):
        message = "HOST HIERARCHY LOOKUP \n HOST: %s has hierarchies:\n\t%s"
        host_name = kwargs.get("host_name")
        # print hierarchy a line per element
        logging.info(message % (host_name, "\n\t".join(str(hiera) for hiera in data.data["hierarchy"])))
        return data


class PrintHierarchyPropertyFilePaths:
    @staticmethod
    def run(data, **kwargs):
        logging.info("HIERARCHY FILES FOUND")
        # print hierarchy a line per element
        for file_name, file_data in data.data["file_hiera"].items():
            message = "Property File %s found: %i references" % (file_name, len(file_data))
            if len(file_data):
                message += "\tMatches: \n\t%s" % ("\n\t".join(str(file_str) for file_str in file_data))
                logging.info(message)
            else:
                logging.debug(message)
        return data


class PrintPropertyFileContents:
    @staticmethod
    def run(data, **kwargs):
        application_name = kwargs.get("application_name")
        host_name = kwargs.get("host_name")
        print("Resolved property files: for %s on host: %s" % (application_name, host_name))

        for file_name, properties in data.data["file_data"].items():
            print("FILE: %s.properties" % file_name)
            for property in properties:
                print("\t", property[0], "=", property[1])
        return data


# -*- coding: utf-8 -*-

from .LoadDataFromDisk import LoadDataFromDisk
import yaml
import os
import glob
import re
import configparser
from itertools import chain
import logging


class LoadHosts(LoadDataFromDisk):
    def merge_contents_of_files(self, file_list):
        contents = dict()
        all_yaml_individual = map(lambda x: yaml.load_all(open(x)), file_list)
        for configs in all_yaml_individual:
            for config in configs:
                contents = {**contents, **config}
        return contents

    def load_contents_of_files(self, glob_pattern=None):
        file_list = self.get_list_of_files(glob_pattern=glob_pattern, recursive=False)
        content = self.merge_contents_of_files(file_list=file_list)
        return content

    def run(self, data, conf, **kwargs):
        base_dir = conf["conf"]["properties"]["base_dir"]
        host_glob = conf["conf"]["properties"]["host_glob"]
        glob_pattern = os.path.join(base_dir, host_glob)
        data.data["hosts"] = self.load_contents_of_files(glob_pattern=glob_pattern)
        data.targets["hosts"] = list(data.data["hosts"].keys())
        return data


class LoadApplicationHostMapping(LoadDataFromDisk):
    def merge_contents_of_files(self, file_list):
        contents = dict()
        for file_path in file_list:
            app_name = file_path.rsplit("/", 1)[1][:-5]
            for config in yaml.load_all(open(file_path)):
                contents[app_name] = config
        return contents

    def load_contents_of_files(self, glob_pattern=None):
        file_list = self.get_list_of_files(glob_pattern=glob_pattern, recursive=False)
        content = self.merge_contents_of_files(file_list=file_list)
        return content

    def run(self, data, conf, **kwargs):
        base_dir = conf["conf"]["properties"]["base_dir"]
        host_mapping_glob = conf["conf"]["properties"]["host_mapping_glob"]
        glob_pattern = os.path.join(base_dir, host_mapping_glob)
        data.data["application_hosts"] = self.load_contents_of_files(glob_pattern=glob_pattern)
        data.targets["applications"] = set(data.data["application_hosts"].keys())
        return data


class LoadApplicationPropertiesMapping(LoadDataFromDisk):
    def merge_contents_of_files(self, file_list):
        contents = dict()
        for file_path in file_list:
            app_name = file_path.rsplit("/", 3)[1]
            for config in yaml.load_all(open(file_path)):
                contents[app_name] = config
        return contents

    def load_contents_of_files(self, glob_pattern=None):
        file_list = self.get_list_of_files(glob_pattern=glob_pattern, recursive=False)
        content = self.merge_contents_of_files(file_list=file_list)
        return content

    def run(self, data, conf, **kwargs):
        base_dir = conf["conf"]["properties"]["base_dir"]
        property_mapping_glob = conf["conf"]["properties"]["property_mapping_glob"]
        glob_pattern = os.path.join(base_dir, property_mapping_glob)
        data.data["application_properties"] = self.load_contents_of_files(glob_pattern=glob_pattern)
        data.targets["applications"] = set(data.data["application_properties"].keys())
        return data


class LoadApplications(LoadDataFromDisk):
    def merge_contents_of_files(self, file_list):
        contents = dict()
        for file_path in file_list:
            app_name = file_path.rsplit("/", 1)[1][:-5]
            for config in yaml.load_all(open(file_path)):
                contents[app_name] = config
        return contents

    def load_contents_of_files(self, base_directory="."):
        glob_pattern = os.path.join(base_directory, "applications", "*.yaml")
        file_list = self.get_list_of_files(glob_pattern=glob_pattern, recursive=False)
        content = self.merge_contents_of_files(file_list=file_list)
        return content


class LoadPropertyFilePaths:

    @staticmethod
    def run(data, conf, **kwargs):
        base_dir = conf["conf"]["properties"]["base_dir"]
        hierarchy = data.data["hierarchy"]
        file_list = dict()
        for hiera in hierarchy:
            hiera_path = hiera["glob"]
            for file_path in glob.iglob(os.path.join(base_dir, "data", hiera_path, '*.properties'), recursive=False):
                file_name = file_path.rsplit("/", 1)[1][:-11]
                if file_name not in file_list:
                    file_list[file_name] = list()
                if "${" in hiera["hiera"]:
                    tags = re.findall(hiera["regex"], file_path)
                else:
                    tags = []
                file_list[file_name].append({"hiera": hiera["hiera"],
                                             "path": file_path,
                                             "tags": tags})

        data.data["file_hiera"] = file_list
        return data


class LoadPropertyFileContents:

    @staticmethod
    def run(data, conf, **kwargs):
        file_list = data.data["file_hiera"]
        resolved_data = dict()
        for file_name, file_path_list in file_list.items():
            parser = configparser.ConfigParser(allow_no_value=True)
            for file_path in file_path_list:
                with open(file_path["path"]) as lines:
                    lines = chain(("[default]",), lines)  # This line does the trick...
                    parser.read_file(lines)

            contents = parser.items("default")
            resolved_data[file_name] = contents
        data.data["file_data"] = resolved_data
        return data


# -*- coding: utf-8 -*-

from .LoadDataFromDisk import LoadDataFromDisk
import yaml
import os
import re
import logging


class LoadHierarchy(LoadDataFromDisk):

    @staticmethod
    def get_regex_hierarchy(hierarchy):
        regex_hierarchy = []
        for hiera in hierarchy:
            regex_hiera = dict()
            regex_hiera['hiera'] = hiera
            regex_hiera['regex'] = os.path.join(re.sub("\${.+?}", "([^/]+)", hiera))
            regex_hiera['tags'] = re.findall("\$\{([^/]+)\}", hiera)
            regex_hiera['glob'] = os.path.join(re.sub("\${.+?}", "*", hiera))
            regex_hierarchy.append(regex_hiera)
        return regex_hierarchy

    def merge_contents_of_files(self, file_list):
        contents = list()
        all_yaml_individual = map(lambda x: yaml.load_all(open(x)), file_list)
        for configs in all_yaml_individual:
            for config in configs:
                contents.extend(config)
        return contents

    def load_contents_of_files(self, glob_pattern=None):
        file_list = self.get_list_of_files(glob_pattern=glob_pattern, recursive=False)
        content = self.merge_contents_of_files(file_list=file_list)
        return content

    def run(self, data, conf, **kwargs):
        base_dir = conf["conf"]["hierarchy"]["base_dir"]
        property_mapping_glob = conf["conf"]["hierarchy"]["hierarchy_glob"]
        glob_pattern = os.path.join(base_dir, property_mapping_glob)
        data.data["hierarchy"] = self.get_regex_hierarchy(self.load_contents_of_files(glob_pattern=glob_pattern))
        return data



# -*- coding: utf-8 -*-
import re


def find_dict_diff(d1, d2, path=""):
    result = False
    for k in d1.keys():
        if k not in d2:
            continue
        elif isinstance(d1[k], dict):
            if path == "":
                path = k
            else:
                path = path + "->" + k
            result = find_dict_diff(d1[k], d2[k], path)
        elif d1[k] == d2[k] or \
                (isinstance(d1[k], list)
                 and isinstance(d2[k], list)
                 and len(set(d1[k]).intersection(set(d2[k]))) > 0):
            return True
    return result


class FilterHostsByHost:
    @staticmethod
    def run(data, conf, **kwargs):
        data_filter = kwargs.get("host_name")
        filtered_targets = list()
        if data_filter in data.targets["hosts"]:
            filtered_targets.append(data_filter)
        data.targets["hosts"] = filtered_targets
        return data


class FilterHostsByTag:
    @staticmethod
    def run(data, **kwargs):
        tag_name = kwargs.get("tag_name")
        tag_value = kwargs.get("tag_value")
        filtered_targets = list()
        for host_name in data.targets["hosts"]:
            host_tags = data.data["hosts"][host_name]
            if tag_name in host_tags and (
                        (isinstance(host_tags[tag_name], str) and host_tags[tag_name] == tag_value) or
                        (isinstance(host_tags[tag_name], list) and any(
                                host_value == tag_value for host_value in host_tags[tag_name]))):
                filtered_targets.append(host_name)
        data.targets["hosts"] = filtered_targets
        return data


class FilterHostsByApplication:
    @staticmethod
    def run(data, **kwargs):
        application_name = kwargs.get("application_name")
        filtered_data = list()
        if application_name in data.targets["applications"]:
            application_tags = data.data["application_hosts"][application_name]
            for host_name in data.targets["hosts"]:
                host_tags = data.data["hosts"][host_name]
                host_app_tag_match = find_dict_diff(host_tags, application_tags)
                if host_app_tag_match:
                    filtered_data.append(host_name)
        data.targets["hosts"] = filtered_data
        return data


class FilterApplicationsByTag:
    @staticmethod
    def run(data, **kwargs):
        tag_name = kwargs.get("tag_name")
        tag_value = kwargs.get("tag_value")
        filtered_targets = list()
        for application_name in data.targets["applications"]:
            application_tags = data.data["application_hosts"][application_name]
            if tag_name in application_tags and (
                        (isinstance(application_tags[tag_name], str)
                         and application_tags[tag_name] == tag_value) or
                        (isinstance(application_tags[tag_name], list)
                         and any(app_value == tag_value for app_value in application_tags[tag_name]))):
                filtered_targets.append(application_name)
        data.targets["applications"] = filtered_targets
        return data


class FilterApplicationsByHost:
    @staticmethod
    def run(data, **kwargs):
        host_name = kwargs.get("host_name")
        filtered_data = list()
        if host_name in data.data["hosts"]:
            host_tags = data.data["hosts"][host_name]
            for application_name in data.targets["applications"]:
                application_tags = data.data["application_hosts"][application_name]
                app_host_tag_match = find_dict_diff(application_tags, host_tags)
                if app_host_tag_match:
                    filtered_data.append(application_name)
        data.targets["applications"] = filtered_data
        return data


class FilterApplicationsByApplication:
    @staticmethod
    def run(data, **kwargs):
        data_filter = kwargs.get("application_name")
        filtered_targets = list()
        if data_filter in data.targets["applications"]:
            filtered_targets.append(data_filter)
        data.targets["applications"] = filtered_targets
        return data


class FilterApplicationsByProperty:
    @staticmethod
    def run(data, **kwargs):
        property_name = kwargs.get("property_name")
        filtered_targets = list()
        for application_name in data.targets["applications"]:
            application_props = data.data["application_properties"][application_name]
            if "properties" in application_props and \
                    isinstance(application_props["properties"], list) and \
                            property_name in application_props["properties"]:
                filtered_targets.append(application_name)
        data.targets["applications"] = filtered_targets
        return data


class FilterHierarchyByHost:
    @staticmethod
    def run(data, **kwargs):
        host_name = kwargs.get("host_name")
        filtered_data = list()
        if host_name in data.data["hosts"]:
            host_tags = data.data["hosts"][host_name].keys()
            for hiera in data.data["hierarchy"]:
                hiera_tags = hiera["tags"]
                if all(hiera_tag in host_tags or hiera_tag == "app" for hiera_tag in hiera_tags):
                    filtered_data.append(hiera)
        data.data["hierarchy"] = filtered_data
        return data


class FilterPropertyFilesByApplication:
    @staticmethod
    def run(data, **kwargs):
        application_name = kwargs.get("application_name")
        filtered_data = data.data["file_hiera"].copy()
        if application_name in data.data["application_properties"]:
            property_files = data.data["application_properties"][application_name]
            for hiera_file in data.data["file_hiera"].keys():
                if hiera_file not in property_files["properties"]:
                    del filtered_data[hiera_file]
        data.data["file_hiera"] = filtered_data
        return data


class FilterPropertyFilesByHostApplicationTags:
    @staticmethod
    def run(data, **kwargs):
        host_name = kwargs.get("host_name")
        application_name = kwargs.get("application_name")
        filtered_data = data.data["file_hiera"].copy()
        if host_name in data.data["hosts"]:
            host_tags = data.data["hosts"][host_name]
            # for each file in list of properties
            # check that path matches
            for file_name, file_datas in data.data["file_hiera"].items():
                for file_data in file_datas:
                    for hiera in data.data["hierarchy"]:
                        if file_data["hiera"] == hiera["hiera"] and len(hiera["tags"]) == len(file_data["tags"]):
                            for i, tag in enumerate(hiera["tags"]):
                                if (tag == "app" and file_data["tags"][i] != application_name) \
                                        or (tag in hiera["tags"] and file_data["tags"][i] in host_tags[tag]):
                                    filtered_data[file_name].remove(file_data)
        data.data["file_hiera"] = filtered_data
        return data


# -*- coding: utf-8 -*-

import os
import logging
import sys


class WriteHostApplicationPropertiesToFiles:
    @staticmethod
    def run(data, conf, **kwargs):
        host_name = kwargs.get("host_name")
        application_name = kwargs.get("application_name")
        output_dir = conf["conf"]["properties"]["output_dir"]
        logging.info("Generating files for application %s on host %s" % (application_name, host_name))
        output_folder = os.path.join(output_dir, host_name, application_name)
        if not os.path.isdir(output_folder):
            os.makedirs(output_folder)

        for file_name, properties in data.data["file_data"].items():
            logging.info("Writing File: %s.properties" % file_name)
            for property in properties:
                with open(os.path.join(output_folder, file_name+".properties"), 'w+') as out:
                    out.write("%s = %s" % (property[0], property[1]))

        return data
